@@ -117,9 +117,7 @@ void ExpressionVisitor::visitAttribute(AttributeAst* node)
 
 void ExpressionVisitor::visitCall(CallAst* node)
 {
-    foreach ( ExpressionAst* c, node->arguments ) {
-        AstDefaultVisitor::visitNode(c);
-    }
+    visitNodeList(node->arguments);
     ExpressionVisitor v(this);
     v.visitNode(node->function);
     auto declaration = Helper::resolveAliasDeclaration(v.lastDeclaration().data());
@@ -277,22 +275,23 @@ AbstractType::Ptr ExpressionVisitor::docstringTypeOverride(
         if ( ! v.lastType() ) {
             return false;
         }
-        ListType::Ptr realTarget;
-        if ( auto target = ListType::Ptr::dynamicCast(normalType) ) {
-            realTarget = target;
+        AbstractType::Ptr newType;
+        if ( auto targetContainer = ListType::Ptr::dynamicCast(normalType) ) {
+            // Copy the return type, to set contents for this call only.
+            docstringType = AbstractType::Ptr(targetContainer->clone());
+            // Add content type of the source.
+            auto sourceContentType = Helper::contentOfIterable(v.lastType(), topContext());
+            ListType::Ptr::staticCast(docstringType)->addContentType<Python::UnsureType>(sourceContentType);
         }
-        if ( auto source = ListType::Ptr::dynamicCast(v.lastType()) ) {
-            if ( ! realTarget ) {
-                // if the function does not force a return type, just copy the source (like for reversed())
-                realTarget = source;
-            }
-            auto newType = ListType::Ptr::staticCast(AbstractType::Ptr(realTarget->clone()));
-            Q_ASSERT(newType);
-            newType->addContentType<Python::UnsureType>(source->contentType().abstractType());
-            docstringType = AbstractType::Ptr::staticCast(newType);
-            return true;
+        else if ( auto sourceContainer = ListType::Ptr::dynamicCast(v.lastType()) ) {
+            // if the function does not force a return type, just copy the source (like for reversed())
+            docstringType = AbstractType::Ptr(sourceContainer->clone());
         }
-        return false;
+        else {
+            return false; // No target container type
+        }
+        Q_ASSERT(docstringType);
+        return true;
     };
 
     foreach ( const QString& currentHint, knownDocstringHints.keys() ) {
@@ -309,6 +308,30 @@ AbstractType::Ptr ExpressionVisitor::docstringTypeOverride(
     return docstringType;
 }
 
+long integerValue(ExpressionAst* node, long wrapTo=0) {
+    bool invert = false;
+    if ( node->astType == Ast::UnaryOperationAstType ) {
+        auto unaryOp = static_cast<UnaryOperationAst*>(node);
+        if ( unaryOp->type == Ast::UnaryOperatorSub ) {
+            node = unaryOp->operand;
+            invert = true;
+        }
+    }
+    if ( node->astType == Ast::NumberAstType ) {
+        auto number = static_cast<NumberAst*>(node);
+        if ( number->isInt ) {
+            // Clamp in case of `a[-9999999:9999999]` or similar.
+            // -1 is just as out-of-range as -99999999, but doesn't cause a huge loop.
+            long upperBound = wrapTo ? wrapTo : LONG_MAX;
+            if (invert) {
+                return qBound(-1L, wrapTo - number->value, upperBound);
+            }
+            return qBound(-1L, number->value, upperBound);
+        }
+    }
+    return LONG_MIN;
+}
+
 void ExpressionVisitor::visitSubscript(SubscriptAst* node)
 {
     AstDefaultVisitor::visitNode(node->value);
@@ -317,40 +340,45 @@ void ExpressionVisitor::visitSubscript(SubscriptAst* node)
     AbstractType::Ptr result(new IntegralType(IntegralType::TypeMixed));
 
     foreach (const auto& type, valueTypes) {
-        if ( (node->slice->astType != Ast::IndexAstType) &&
-             (type.cast<IndexedContainer>() || type.cast<ListType>()) ) {
+        if ( node->slice->astType == Ast::SliceAstType ) {
+            auto slice = static_cast<SliceAst*>(node->slice);
+            if ( auto tupleType = type.cast<IndexedContainer>() ) {
+                DUChainReadLocker lock;
+                auto newTuple = typeObjectForIntegralType<IndexedContainer>("tuple");
+                if ( ! newTuple ) {
+                    continue;
+                }
+                long step = slice->step ? integerValue(slice->step) : 1;
+                int len = tupleType->typesCount();
+                long lower = slice->lower ? integerValue(slice->lower, len) : ((step > 0) ? 0 : len - 1);
+                long upper = slice->upper ? integerValue(slice->upper, len) : ((step > 0) ? len : -1);
+                if ( step != LONG_MIN && lower != LONG_MIN && upper != LONG_MIN) {
+                    long ii = lower;
+                    while ( (upper - ii) * step > 0 ) {
+                        if ( 0 <= ii && ii < len ) {
+                            newTuple->addEntry(tupleType->typeAt(ii).abstractType());
+                        }
+                        ii += step;
+                    }
+                }
+                result = Helper::mergeTypes(result, newTuple);
+                continue;
+            }
+        }
+        if ( (node->slice->astType != Ast::IndexAstType) && type.cast<ListType>() ) {
             if ( type.cast<MapType>() ) {
                 continue; // Can't slice dicts.
             }
-            // Assume that slicing (e.g. foo[3:5]) a tuple/list returns the same type.
-            // TODO: we could do better for some tuple slices.
+            // Assume that slicing (e.g. foo[3:5]) a list returns the same type.
             result = Helper::mergeTypes(result, type);
         }
         else if ( const auto& indexed = type.cast<IndexedContainer>() ) {
-            IndexAst* sliceIndexAst = static_cast<IndexAst*>(node->slice);
-            NumberAst* number = nullptr;
-            bool invert = false;
-            if ( sliceIndexAst->value->astType == Ast::UnaryOperationAstType ) {
-                // might be -3
-                UnaryOperationAst* unary = static_cast<UnaryOperationAst*>(sliceIndexAst->value);
-                if ( unary->type == Ast::UnaryOperatorSub && unary->operand->astType == Ast::NumberAstType ) {
-                    number = static_cast<NumberAst*>(unary->operand);
-                    invert = true;
-                }
-            }
-            else if ( sliceIndexAst->value->astType == Ast::NumberAstType ) {
-                number = static_cast<NumberAst*>(sliceIndexAst->value);
-            }
-            if ( number ) {
-                int sliceIndex = number->value * ( invert ? -1 : 1 );
-                if ( sliceIndex < 0 && sliceIndex + indexed->typesCount() >= 0 ) {
-                    sliceIndex += indexed->typesCount();
-                }
-                if ( sliceIndex >= 0 && sliceIndex < indexed->typesCount() ) {
-                    result = Helper::mergeTypes(result, indexed->typeAt(sliceIndex).abstractType());
-                    continue;
-                }
+            long sliceIndex = integerValue(static_cast<IndexAst*>(node->slice)->value, indexed->typesCount());
+            if ( 0 <= sliceIndex && sliceIndex < indexed->typesCount() ) {
+                result = Helper::mergeTypes(result, indexed->typeAt(sliceIndex).abstractType());
+                continue;
             }
+            // Index is unknown or invalid, could be returning any of the content types.
             result = Helper::mergeTypes(result, indexed->asUnsureType());
         }
         else if ( const auto& listType = type.cast<ListType>() ) {
@@ -586,6 +614,19 @@ void ExpressionVisitor::visitBytes(Python::BytesAst* ) {
     encounter(AbstractType::Ptr::staticCast(type));
 }
 
+void ExpressionVisitor::visitFormattedValue(Python::FormattedValueAst* ) {
+    DUChainReadLocker lock;
+    StructureType::Ptr type = typeObjectForIntegralType<StructureType>("str");
+    encounter(AbstractType::Ptr::staticCast(type));
+}
+
+void ExpressionVisitor::visitJoinedString(Python::JoinedStringAst* )
+{
+    DUChainReadLocker lock;
+    StructureType::Ptr type = typeObjectForIntegralType<StructureType>("str");
+    encounter(AbstractType::Ptr::staticCast(type));
+}
+
 RangeInRevision nodeRange(Python::Ast* node)
 {
     return RangeInRevision(node->startLine, node->startCol, node->endLine,node->endCol);
@@ -612,19 +653,19 @@ void ExpressionVisitor::visitNameConstant(NameConstantAst* node)
 
 void ExpressionVisitor::visitName(Python::NameAst* node)
 {
-    RangeInRevision range;
+    CursorInRevision findNameBefore;
     if ( m_scanUntilCursor.isValid() ) {
-        range = RangeInRevision(CursorInRevision(0, 0), m_scanUntilCursor);
+        findNameBefore = m_scanUntilCursor;
     }
     else if ( m_forceGlobalSearching ) {
-        range = RangeInRevision::invalid();
+        findNameBefore = CursorInRevision::invalid();
     }
     else {
-        range = RangeInRevision(0, 0, node->endLine, node->endCol);
+        findNameBefore = CursorInRevision(node->endLine, node->endCol);
     }
     DUChainReadLocker lock;
-    Declaration* d = Helper::declarationForName(QualifiedIdentifier(node->identifier->value),
-                                                range, DUChainPointer<const DUContext>(context()));
+    Declaration* d = Helper::declarationForName(node, findNameBefore,
+                                                DUChainPointer<const DUContext>(context()));
 
     if ( d ) {
         bool isAlias = dynamic_cast<AliasDeclaration*>(d) || d->isFunctionDeclaration() || dynamic_cast<ClassDeclaration*>(d);
@@ -705,11 +746,19 @@ void ExpressionVisitor::visitUnaryOperation(Python::UnaryOperationAst* node)
 
 void ExpressionVisitor::visitBooleanOperation(Python::BooleanOperationAst* node)
 {
-    foreach (ExpressionAst* expression, node->values) {
-        visitNode(expression);
+    ExpressionVisitor v(this);
+    AbstractType::Ptr result;
+
+    for (const auto& expr : node->values) {
+        v.visitNode(expr);
+        result = Helper::mergeTypes(result, v.lastType());
     }
 
-    encounter(AbstractType::Ptr(new IntegralType(IntegralType::TypeBoolean)));
+    encounter(result);
+}
+
+void ExpressionVisitor::visitAssignmentExpression(Python::AssignmentExpressionAst* node) {
+    visitNode(node->value);
 }
 
 }
